/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */


#ifdef WIN32 

#include <stdio.h>
#include <iostream>
#include <map>
#include <windows.h>
#include <Producer/RenderSurface>
#include <Producer/Events>
#include <Producer/Keyboard>

// Use the GLUT CURSOR defines for now
#define  GLUT_CURSOR_RIGHT_ARROW            0x0000
#define  GLUT_CURSOR_LEFT_ARROW             0x0001
#define  GLUT_CURSOR_INFO                   0x0002
#define  GLUT_CURSOR_DESTROY                0x0003
#define  GLUT_CURSOR_HELP                   0x0004
#define  GLUT_CURSOR_CYCLE                  0x0005
#define  GLUT_CURSOR_SPRAY                  0x0006
#define  GLUT_CURSOR_WAIT                   0x0007
#define  GLUT_CURSOR_TEXT                   0x0008
#define  GLUT_CURSOR_CROSSHAIR              0x0009
#define  GLUT_CURSOR_UP_DOWN                0x000A
#define  GLUT_CURSOR_LEFT_RIGHT             0x000B
#define  GLUT_CURSOR_TOP_SIDE               0x000C
#define  GLUT_CURSOR_BOTTOM_SIDE            0x000D
#define  GLUT_CURSOR_LEFT_SIDE              0x000E
#define  GLUT_CURSOR_RIGHT_SIDE             0x000F
#define  GLUT_CURSOR_TOP_LEFT_CORNER        0x0010
#define  GLUT_CURSOR_TOP_RIGHT_CORNER       0x0011
#define  GLUT_CURSOR_BOTTOM_RIGHT_CORNER    0x0012
#define  GLUT_CURSOR_BOTTOM_LEFT_CORNER     0x0013
#define  GLUT_CURSOR_INHERIT                0x0064
#define  GLUT_CURSOR_NONE                   0x0065
#define  GLUT_CURSOR_FULL_CROSSHAIR         0x0066

    // Set the cursor AND change it for this window class. 
#define MAP_CURSOR(a,b) case a: SetCursor( LoadCursor(NULL,b) ); \
SetClassLong(_win,GCL_HCURSOR,LONG(LoadCursor(NULL,b))); \
break;
    // Nuke the cursor AND change it for this window class. 
#define ZAP_CURSOR(a,b) case a: SetCursor(NULL); \
SetClassLong(_win,GCL_HCURSOR,LONG(NULL)); \
break;


using namespace Producer;

// Class statics
std::map <Window, RenderSurface *>RenderSurface::registry;


RenderSurface::Client::Client(unsigned long mask): _mask(mask)
{
        q = new EventQueue;
}

void RenderSurface::Client::queue( ref_ptr<Event> ev ) 
{ 
        q->push_back( ev ); 
}


LONG WINAPI RenderSurface::s_proc( Window hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    std::map <Window, RenderSurface *>::iterator p;
    p = registry.find( hWnd );
    if( p == registry.end() )
        return DefWindowProc( hWnd, uMsg, wParam, lParam );
    else
        return p->second->proc( hWnd, uMsg, wParam, lParam );
        
}

void RenderSurface::dispatch( ref_ptr<Event> ev )
{
    if( clients.size() == 0 ) return;

        
        std::vector <Producer::ref_ptr<Client> >::iterator p;
    for( p = clients.begin(); p != clients.end(); p++ )
    {
#if 0
        if( ev->type() == Event::MouseMotion )
        {
            // If MouseMotionMask is set, then send event regardless of
            // Mouse button state.
            if( (*p)->mask() & Event::MouseMotionMask )
                (*p)->queue(ev);
            // Otherwise, send event only if mouse button is pressed
            else if( (*p)->mask() & Event::ButtonMotionMask && _mbutton )
                (*p)->queue(ev);
        }
        else if( (*p)->mask() & (1<<ev->type()) )
            (*p)->queue( ev ); 
#endif

        if( (*p)->mask() & (1<<ev->type()) )
                {                
            (*p)->queue( ev ); 
                }
    } 
}


LONG WINAPI RenderSurface::proc( Window hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    LONG lRet = 1;
    RECT rect;
    ref_ptr<Event> ev;

    switch (uMsg)
    {
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 522
#endif
        case WM_MOUSEWHEEL:
                {
                        KeyboardMouseCallback::ScrollingMotion scroll = 
                                (signed int)(wParam) < 0 ? KeyboardMouseCallback::ScrollDown:
                                KeyboardMouseCallback::ScrollUp;
                        ev = new MouseScrollEvent(hWnd,scroll);
                        dispatch(ev);
                }
                if (!_ownWindow)
                        lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
                break;
    case WM_MOUSEMOVE :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));

        ev = new MouseMotionEvent( hWnd, _mx, _my );
        dispatch( ev );
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_LBUTTONDOWN :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new ButtonPressEvent( hWnd, 1, _mx, _my );
        dispatch( ev );
        _mbutton |= 1L;
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_LBUTTONUP :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));

        ev =  new ButtonReleaseEvent( hWnd, 1, _mx, _my );
        dispatch( ev );
        _mbutton &= ~1L;
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_LBUTTONDBLCLK :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new DoubleButtonPressEvent( hWnd, 1, _mx, _my );
        dispatch( ev );
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_RBUTTONDOWN :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new ButtonPressEvent( hWnd, 3, _mx, _my );
        dispatch( ev );
        _mbutton |= 4L;
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_RBUTTONUP  :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new ButtonReleaseEvent( hWnd, 3, _mx, _my );
        dispatch( ev );
        _mbutton &= ~4L;
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_RBUTTONDBLCLK : 
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        dispatch( new DoubleButtonPressEvent( hWnd, 3, _mx, _my ));
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_MBUTTONDOWN  :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new ButtonPressEvent( hWnd, 2, _mx, _my );
        dispatch( ev );
        _mbutton |= 2L;
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_MBUTTONUP :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new ButtonReleaseEvent( hWnd, 2, _mx, _my );
        dispatch( ev );
        _mbutton &= ~2L;
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_MBUTTONDBLCLK :
        _mx = ((int)(short)LOWORD(lParam));
        _my = ((int)(short)HIWORD(lParam));
        ev = new DoubleButtonPressEvent( hWnd, 2, _mx, _my );
        dispatch( ev );
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_PAINT:
        if (_ownWindow)
        {
            PAINTSTRUCT    ps;
            BeginPaint(hWnd, &ps);
            EndPaint(hWnd, &ps);
        }
        else
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_CREATE:
    case WM_SIZE:
    case WM_MOVE:

        // only change window dimensions for non-child, 
        // non-fullscreen windows (which are all that 
        // should be moved or resized directly.
        if (!(_isFullScreen || _parent))
        {
            GetClientRect(hWnd, &rect);
            _windowWidth = rect.right - rect.left;
            _windowHeight = rect.bottom - rect.top;

            // map the window origin to the screen
            POINT p;
            p.x = rect.left;
            p.y = rect.top;
            ClientToScreen(hWnd, &p);

            // flip to Y up convention
            _windowX = p.x;
            _windowY = _screenHeight - p.y - _windowHeight;

        }

        ev = new WindowConfigEvent( hWnd, _windowX, _windowY,
                                        _windowWidth,
                                        _windowHeight);
        dispatch(ev);
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_CLOSE:
        if (!_ownWindow)
        {
			// PROBLEM: can't call KillGLWindow() here because the GL context
			// was created in another thread and releasing it now would cause 
			// an error!
            //KillGLWindow();
            _realized = false;
            SetWindowLong(_win, GWL_WNDPROC, (long)_oldWndProc);
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
            ev = new ShutdownEvent(hWnd);
            dispatch(ev);
        }
        else
            DestroyWindow (hWnd);                
        break;

    case WM_QUIT:
    case WM_DESTROY:
        if (!_ownWindow)
        {
			// PROBLEM: can't call KillGLWindow() here because the GL context
			// was created in another thread and releasing it now would cause 
			// an error!
            //KillGLWindow();
            _realized = false;
            SetWindowLong(_win, GWL_WNDPROC, (long)_oldWndProc);
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        }
        else
        {
            PostQuitMessage (0);
            _realized = false;
        }
        ev = new ShutdownEvent(hWnd);
        dispatch(ev);
        break;

    case WM_SYSKEYDOWN:
    case WM_KEYDOWN:
        {
            KeyCharacter keychar = Keyboard::getSingleton()->win32KeyEventToKeyChar( lParam, wParam );
            ev = new KeyPressEvent(hWnd, keychar);
            dispatch( ev );
        }
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_SYSKEYUP:
    case WM_KEYUP:
        {
            KeyCharacter keychar = Keyboard::getSingleton()->win32KeyEventToKeyChar( lParam, wParam );
            ev = new KeyReleaseEvent(hWnd, keychar);
            dispatch( ev );
        }
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;
        /*
    case WM_CHAR:
        if (lParam >> 31) {     // this bit will be set if this event is a key release
            ev = new KeyReleaseEvent(wParam);
        } else {
            ev = new KeyPressEvent(wParam);
        }
        dispatch( ev );

        if( _use_default_esc && wParam == VK_ESCAPE )
        {
            PostQuitMessage(0);
            return 0;
        }
        */

    case WM_TIMER:
        ev = new TimerEvent(hWnd, wParam);
        dispatch( ev );
        if (!_ownWindow)
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;

    case WM_SETCURSOR:
    {
        long lRet = TRUE;
        if (LOWORD(lParam) == HTCLIENT)
        {
            if (!_ownWindow)
                lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
            _setCursor();
        }
        else
        {
            if (_ownWindow)
                lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
            else
                lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        }
        break;
    }

    default:
        if (_ownWindow)
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        else
            lRet = CallWindowProc(_oldWndProc, hWnd, uMsg, wParam, lParam);
        break;
    }
    return lRet;
} 

bool RenderSurface::realize( VisualChooser *vc, GLContext )
{
    if( _realized ) return _realized;
    if( vc != NULL )
        _visualChooser = vc;
    if( _useConfigEventThread && _ownWindow )
    {
        _threadReady = new RefBarrier(2);
        startThread();
        _threadReady->block();
    }
    else
        _init();
        
    makeCurrent();

    _realized = true;

        std::vector <Producer::ref_ptr<Callback> >::iterator p;
    for( p = _realizeCallbacks.begin(); p != _realizeCallbacks.end(); p++ )
    {
        if( (*p).valid() )
            (*(*p).get())( *this );
    }
    _realizeCallbacks.clear();

    _realizeBlock->release();

    return _realized;
}


static void doInternalError( char *msg )
{

    LPVOID lpMsgBuf;
    FormatMessage(   FORMAT_MESSAGE_ALLOCATE_BUFFER |
                     FORMAT_MESSAGE_FROM_SYSTEM |
                     FORMAT_MESSAGE_IGNORE_INSERTS,
                     NULL,
                     GetLastError(),
                     MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                     (LPTSTR) &lpMsgBuf,\
                     0,NULL 
                     );
    char *szMessage = new char[strlen(msg)+5+strlen("Producer: Internal Error \n ")+2];
    sprintf( szMessage, "Producer: Internal Error \n %s", msg );
    MessageBox( NULL, (LPCTSTR)lpMsgBuf, szMessage, MB_OK | MB_ICONINFORMATION );
    LocalFree( lpMsgBuf );
    delete szMessage;
}


void RenderSurface::KillGLWindow()
{
    if (_glcontext)
    {
        makeCurrent();
    }

    if (_isFullScreen)
    {
        // Switch Back To The Desktop
        ChangeDisplaySettings(NULL,0);    
        ShowCursor(TRUE);
    }

    if (_glcontext && !wglDeleteContext(_glcontext))
    {
        doInternalError("Release Rendering Context Failed.");
    }
    _glcontext = NULL;

    if (_hdc && !ReleaseDC(_win,_hdc))
    {
        doInternalError("Release Device Context Failed.");
    }
    _hdc = NULL;

    if (_ownWindow)
    {
        if (_win && !DestroyWindow(_win))
        {
            doInternalError("Could Not Release hWnd.");
        }
        _win = NULL;
        if (!UnregisterClass(_windowName.c_str(),_hinstance))
        {
            doInternalError("Could Not Unregister Class.");
        }
        _hinstance = NULL;
    }
}

bool RenderSurface::_init()
{
    _hinstance    = GetModuleHandle(NULL);
    
    RECT rect;
    DEVMODE dm;
    memset(&dm,0,sizeof(dm));
    dm.dmSize = sizeof(dm);
    EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm);
    int x, y, w, h;

    _screenWidth  = dm.dmPelsWidth;
    _screenHeight = dm.dmPelsHeight;

    _setCursor(GLUT_CURSOR_LEFT_ARROW);

    if (_ownWindow )
    {
        DWORD dwExStyle;
        DWORD dwStyle;
        WNDCLASS wndclass;
        
        wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; 
        wndclass.lpfnWndProc   = (WNDPROC)s_proc;
        wndclass.cbClsExtra    = 0;
        wndclass.cbWndExtra    = 0;
        wndclass.hInstance     = _hinstance;
        wndclass.hCursor       = LoadCursor (NULL,IDC_ARROW);
        wndclass.hIcon         = LoadIcon( _hinstance, "PRODUCER_ICON" );
        wndclass.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
        wndclass.lpszMenuName  = 0;
    
        wndclass.lpszClassName = _windowName.c_str();

        // mew 2003-12-14
        // if the .exe doesn't override the PRODUCER_ICON resource
        // definition, fall back to the icon in the producer .dll.
#ifdef _DEBUG
#define LIBNAME "Producerd.dll"
#else
#define LIBNAME "Producer.dll"
#endif
        if (wndclass.hIcon == NULL)
        {
            HINSTANCE hDLLInstance = LoadLibrary( LIBNAME );
            if (hDLLInstance != NULL)
            {
                wndclass.hIcon = LoadIcon(hDLLInstance, "PRODUCER_ICON");
            }
        }
        RegisterClass(&wndclass);

        if (_parent)
        {
            _isFullScreen = false;
        }

        if (_isFullScreen)
        {
            if (ChangeDisplaySettings(&dm,CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
            {
                // If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
                if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","",MB_YESNO|MB_ICONEXCLAMATION)==IDYES)
                {
                    _isFullScreen = FALSE;
                }
                else
                {
                    doInternalError("Program Will Now Close.");
                    return FALSE;
                }
            }
        }

        if (_parent)
        {
            GetClientRect(_parent, &rect);

            // style is a child window
            dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
            dwStyle   = WS_CHILDWINDOW;

            //AdjustWindowRectEx(&rect, dwStyle, FALSE, dwExStyle);

            _windowWidth = (rect.right - rect.left);
            _windowX = 0;

            _windowHeight = (rect.bottom - rect.top);
            _windowY = 0;

            x = _windowX;
            y = _windowY;
            w = _windowWidth;
            h = _windowHeight;
        }
        else if (_isFullScreen)
        {
            dwExStyle = WS_EX_APPWINDOW;
            dwStyle   = WS_POPUP | WS_MAXIMIZE;

            x = 0;
            y = 0;
            w = _screenWidth;
            h = _screenHeight;
        }
        else // own window, no parent, not fullscreen
        {
            if( _windowWidth == 0xFFFFFFFF ) 
            {
                _windowWidth = _screenWidth/2;
                _windowX = _windowWidth/2;
            } 

            if ( _windowHeight == 0xFFFFFFFF ) 
            {
                _windowHeight = _screenHeight/2;
                _windowY = _windowHeight/2;
            } 
             
            dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
            dwStyle   = WS_OVERLAPPEDWINDOW;

            // map the Y up dimensions to Y down for window
            rect.left = _windowX;
            rect.top = _screenHeight - _windowY - _windowHeight;
            rect.bottom = _screenHeight - _windowY;
            rect.right = _windowWidth + _windowX;

            // map from client area dimensions to outer window dimensions
            AdjustWindowRectEx(&rect, dwStyle, FALSE, dwExStyle);

            x = rect.left;
            y = rect.top;
            w = rect.right - rect.left;
            h = rect.bottom - rect.top;
        }

        if (!(_win = CreateWindowEx(dwExStyle,
                                    _windowName.c_str(),
                                    TEXT( _windowName.c_str() ),
                                    dwStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                                    x,
                                    y,
                                    w,
                                    h,
                                    _parent,
                                    NULL,
                                    _hinstance,
                                    NULL)))
        {
            doInternalError("Window Creation Error.");
            KillGLWindow();
            return FALSE;
        }
    }
    else if( (_windowWidth == 0xFFFFFFFF) && (_windowHeight == 0xFFFFFFFF) )
    {
        POINT p;

        GetClientRect(_win, &rect);
        _windowWidth = rect.right - rect.left;
        _windowHeight = rect.bottom - rect.top;

        // map window origin to screen
        p.x = rect.left;
        p.y = rect.top;
        ClientToScreen(_win, &p);

        _windowX = p.x;
        _windowY = _screenHeight - _windowHeight - p.y;
        _isFullScreen = false;
        _oldWndProc = (WNDPROC)SetWindowLong(_win, GWL_WNDPROC, (long)((WNDPROC)s_proc));
    }

    if (!(_hdc=GetDC(_win)))
    {
        doInternalError("Can't Create A GL Device Context.");
        KillGLWindow();
        return FALSE;
    }

    unsigned int visualId;
    if (_visualChooser == NULL )
    {
        Producer::ref_ptr<VisualChooser> lvc = new Producer::VisualChooser;
        lvc->setSimpleConfiguration();
        lvc->choose(&_hdc, _screen);
        visualId = lvc->getVisualID();
    }
    else
    {
        _visualChooser->choose(&_hdc,_screen);
        visualId = _visualChooser->getVisualID();
    }

    PIXELFORMATDESCRIPTOR pfd;
    if (!SetPixelFormat(_hdc, visualId, &pfd))
    {
        doInternalError("Can't Set The PixelFormat.");
        KillGLWindow();
        return FALSE;
    }

    if (!(_glcontext = wglCreateContext( _hdc )))
    {
        doInternalError("Can't Create A GL Rendering Context.");
        KillGLWindow();
        return FALSE;
    }

// ?? WHY DOES THIS CRASH HERE ??  nhv
//    if (!makeCurrent())
//    {
//        KillGLWindow();
//        MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
//        return FALSE;
//    }


    _setCursor();
    ShowWindow(_win, SW_NORMAL );
    UpdateWindow(_win );
    SetForegroundWindow(_win);
    SetFocus(_win);

    registry.insert(std::pair<Window,RenderSurface *>(_win, this));

    return true;
}


bool RenderSurface::makeCurrent( void )

{
    return(wglMakeCurrent(_hdc, _glcontext));
}

void RenderSurface::fullScreen( bool flag )
{
    int x, y, w, h;

    if( flag == _isFullScreen )
        return;

    if (_parent)
        return;

    _isFullScreen = flag;

    if (_isFullScreen)
    {
        /* force fullscreen mode */
        DWORD s = GetWindowLong(_win, GWL_STYLE);
        s &= ~WS_OVERLAPPEDWINDOW;
        s |= WS_POPUP;
        SetWindowLong(_win, GWL_STYLE, s);
        SetWindowPos(_win, 
                     HWND_TOP,
                     0, 0, 
                     _screenWidth, _screenHeight, 
                     SWP_FRAMECHANGED);

    } 
    else 
    {
        UINT flags = SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOSENDCHANGING |
                     SWP_NOSIZE | SWP_NOZORDER;

        DWORD style = GetWindowLong(_win, GWL_STYLE);

        /* Get rid of fullscreen mode, if it exists */
        if ( style & WS_POPUP ) {
            style &= ~WS_POPUP;
            style |= WS_OVERLAPPEDWINDOW;
            SetWindowLong(_win, GWL_STYLE, style);
            flags |= SWP_FRAMECHANGED;
        }

        flags &= ~SWP_NOMOVE;
        flags &= ~SWP_NOSIZE;

        RECT rect;

        // map Y up convention to Y down for windows
        rect.left = _windowX;
        rect.top = _screenHeight - _windowY - _windowHeight;
        rect.bottom = _screenHeight - _windowY;
        rect.right = _windowWidth + _windowX;

        // adjust from client dimensions to window outer dimentions
        AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, FALSE, WS_EX_APPWINDOW | WS_EX_WINDOWEDGE);

        x = rect.left;
        y = rect.top;
        w = rect.right - rect.left;
        h = rect.bottom - rect.top;

        SetWindowPos(_win,
                     HWND_TOP,
                     x, y, w, h,
                     flags);
    }
}

bool RenderSurface::makeCurrent( void ) const
{
    return(wglMakeCurrent( _hdc, _glcontext ));
}

void RenderSurface::swapBuffers( void )
{
    SwapBuffers( _hdc );
}

EventQueue * RenderSurface::selectInput( unsigned int mask )
{
    Client *client = new Client(mask);
    clients.push_back( client );
    // Need to inform clients that the windows has already
    // been created and what its size is
    if( _realized )
    {        client->queue( new WindowConfigEvent( _win,
            _windowX,
            _windowY,
            _windowWidth,
            _windowHeight ));
    }
    return client->getQueue();
}

void RenderSurface::sync( int divisor )
{
    divisor = divisor;
    glFinish();
}


void RenderSurface::run( void )
{
    _init();
    if( _threadReady.valid())
        _threadReady->block();

    MSG msg;

    while( GetMessage( &msg, NULL, 0, 0) > 0)
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
                testCancel();
    }
}

void RenderSurface::_resizeWindow()
{
    RECT rect;
    int x, y, h, w;

    if (_parent)
    {
        // get the client area of the parent
        GetClientRect(_parent, &rect);

        // clip to width of parent
        if ((_windowWidth + _windowX) < (unsigned int)rect.right)
        {
            w = _windowWidth;
            x = _windowX;
        }
        else
        {
            w = rect.right - _windowX;
            x = _windowX;
        }

        // clip to height of parent and flip from
        // Y up convention to Y down for window
        if ((_windowY + _windowHeight) < (unsigned int)rect.bottom)
        {
            h = _windowHeight;
            y = rect.bottom - _windowY - _windowHeight;
        }
        else
        {
            h = rect.bottom - _windowY;
            y = 0;
        }
    }
    else if (_isFullScreen)
    {
        x = 0;
        y = 0;
        h = _screenHeight;
        w = _screenWidth;
    }
    else // normal window
    {
        RECT rect;

        // map from Y up to Y down
        rect.left = _windowX;
        rect.top = _screenHeight - _windowY - _windowHeight;
        rect.bottom = _screenHeight - _windowY;
        rect.right = _windowWidth + _windowX;

        // adjust from client dimension to outside dimension
        //AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, FALSE, WS_EX_APPWINDOW | WS_EX_WINDOWEDGE);
        // Above line replaced with this contribution by Trajce Nikolov (Nick):
        // Leaving the old line in commented out in case...
        LONG style = GetWindowLong(_win,GWL_STYLE);
        LONG styleEx = GetWindowLong(_win,GWL_EXSTYLE);
        AdjustWindowRectEx(&rect, style, FALSE, styleEx);

        x = rect.left;
        y = rect.top;
        w = rect.right - rect.left;
        h = rect.bottom - rect.top;
    }

    SetWindowPos(_win,HWND_TOP, x, y, w,
                 h, SWP_SHOWWINDOW | SWP_FRAMECHANGED);
}

void RenderSurface::_setBorder( bool flag )
{
    std::cout << "RenderSurface::_setBorder( "<< flag << " ) : Warning - Not implemented yet\n";

}

void RenderSurface::_setCursor()
{
    POINT point;
    GetCursorPos(&point);
    if (_win == WindowFromPoint(point))
    {
        if( _useCursor )
        {
            _setCursor(_curCursor);
        } 
        else 
        {
            _setCursor(GLUT_CURSOR_NONE);
        }
    }
}

void RenderSurface::_setCursor(int cursorID)
{
    /*
     * This is a temporary solution only...
     */
    switch( cursorID )
    {
        MAP_CURSOR( GLUT_CURSOR_RIGHT_ARROW, IDC_ARROW     );
        MAP_CURSOR( GLUT_CURSOR_LEFT_ARROW,  IDC_ARROW     );
        MAP_CURSOR( GLUT_CURSOR_INFO,        IDC_HELP      );
        MAP_CURSOR( GLUT_CURSOR_DESTROY,     IDC_CROSS     );
        MAP_CURSOR( GLUT_CURSOR_HELP,        IDC_HELP       );
        MAP_CURSOR( GLUT_CURSOR_CYCLE,       IDC_SIZEALL   );
        MAP_CURSOR( GLUT_CURSOR_SPRAY,       IDC_CROSS     );
        MAP_CURSOR( GLUT_CURSOR_WAIT,         IDC_WAIT      );
        MAP_CURSOR( GLUT_CURSOR_TEXT,        IDC_UPARROW   );
        MAP_CURSOR( GLUT_CURSOR_CROSSHAIR,   IDC_CROSS     );
        ZAP_CURSOR( GLUT_CURSOR_NONE,        NULL       );

        default:
            MAP_CURSOR( GLUT_CURSOR_UP_DOWN,     IDC_ARROW     );
    }
    /*
     * Remember the currently selected cursor
     */
    if (cursorID != GLUT_CURSOR_NONE)
    {
        _curCursor = cursorID;
    }
}

void RenderSurface::_positionPointer(int x, int y)
{
    //****TODO***
    // This needs to be mapped from an x, y that
    // are interpreted in the Y up Producer world
    // to Y down in the window world.  Should x, y
    // be screen or window coordinates?

    POINT coords = { x, y };

    // First find the new screen-relative coordinates of the mouse
    ClientToScreen( _win, &coords );

    // Now set the new mouse cursor position...
    SetCursorPos( coords.x, coords.y );
}

void RenderSurface::getScreenSize( unsigned int &width, unsigned int &height ) const
{
    if( _realized )
    {
        width = _screenWidth;
        height = _screenHeight;
    }
    else
    {
        DEVMODE dm;
        memset(&dm,0,sizeof(dm));
        dm.dmSize = sizeof(dm);
        EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm);
        width  = dm.dmPelsWidth;
        height = dm.dmPelsHeight;
    }
}

void RenderSurface::getScreenSize( unsigned int &width, unsigned int &height ) 
{
    if( _realized )
    {
        width = _screenWidth;
        height = _screenHeight;
    }
    else
    {
        DEVMODE dm;
        memset(&dm,0,sizeof(dm));
        dm.dmSize = sizeof(dm);
        EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm);
        width  = dm.dmPelsWidth;
        height = dm.dmPelsHeight;
    }
}

void RenderSurface::resizeToParent()
{
    RECT rect;

    if (_parent)
    {
        GetClientRect(_parent, &rect);

        _windowWidth = (rect.right - rect.left);
        _windowX = 0;

        _windowHeight = (rect.bottom - rect.top);
        _windowY = 0;

        _resizeWindow();
    }
}

void RenderSurface::mapWindow()
{
    fprintf( stderr, "RenderSurface::mapWindow() - Unimplemented under win32\n" );
}

void RenderSurface::unmapWindow()
{
    fprintf( stderr, "RenderSurface::unMapWindow() - Unimplemented under win32\n" );
}

void RenderSurface::_fini()
{
    if( _realized )
	{
		if (!_ownWindow)
			SetWindowLong(_win, GWL_WNDPROC, (long)_oldWndProc);
	    PostThreadMessage(getThreadId(), WM_QUIT, 0, 0);
	}
}

   
#endif  // WIN32
