/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */
#include <stdio.h>
#include <iostream>

#include <Producer/Export>
#include <Producer/Referenced>
#include <Producer/VisualChooser>

using namespace std;
using namespace Producer;


VisualChooser::VisualChooser( void )
{
    _visual_id = 0;
    _vinfo = 0L;
}

VisualChooser::~VisualChooser(void)
{
   clear();
}

void VisualChooser::setVisual( VisualInfo *vinfo )
{
    clear();
    _vinfo = vinfo;
}


void VisualChooser::addAttribute( AttributeName attribute )
{
    _vinfo = 0L;
    _visual_attributes.push_back( VisualAttribute( attribute ) );
}

void VisualChooser::addAttribute( AttributeName attribute, int parameter )
{
    _vinfo = 0L;
    _visual_attributes.push_back( VisualAttribute( attribute, parameter ) );
}

void VisualChooser::addExtendedAttribute( unsigned int attribute )
{
    _vinfo = 0L;
    _visual_attributes.push_back( VisualAttribute( attribute ) );
}

void VisualChooser::addExtendedAttribute( unsigned int attribute, int parameter )
{
    _vinfo = 0L;
    _visual_attributes.push_back( VisualAttribute( attribute, parameter ) );
}

void VisualChooser::setBufferSize( unsigned int size )
{
    addAttribute( BufferSize, size );
}

void VisualChooser::setLevel( int level )
{
    addAttribute( Level, level );

}
void VisualChooser::useRGBA()
{
    addAttribute( RGBA );
}
void VisualChooser::useDoubleBuffer()
{
    addAttribute( DoubleBuffer );
}
void VisualChooser::useStereo()
{
    addAttribute( Stereo );
}

void VisualChooser::setAuxBuffers( unsigned int num )
{
    addAttribute( AuxBuffers, num );
}

void VisualChooser::setRedSize( unsigned int size )
{
    addAttribute( RedSize, size );
}

void VisualChooser::setGreenSize( unsigned int size )
{
    addAttribute( GreenSize, size );
}

void VisualChooser::setBlueSize( unsigned int size )
{
    addAttribute( BlueSize, size );
}

void VisualChooser::setAlphaSize( unsigned int size )
{
    addAttribute( AlphaSize, size );
}

void VisualChooser::setDepthSize( unsigned int size )
{
    addAttribute( DepthSize, size );
}

void VisualChooser::setStencilSize( unsigned int size )
{
    addAttribute( StencilSize, size );
}

void VisualChooser::setAccumRedSize( unsigned int size )
{
    addAttribute( AccumRedSize, size );
}

void VisualChooser::setAccumGreenSize( unsigned int size )
{
    addAttribute( AccumGreenSize, size );
}

void VisualChooser::setAccumBlueSize( unsigned int size )
{
    addAttribute( AccumBlueSize, size );
}

void VisualChooser::setAccumAlphaSize( unsigned int size )
{
    addAttribute( AccumAlphaSize, size );
}

void VisualChooser::setVisualID( unsigned int id ) 
{ 
    _visual_id = id; 
}

void VisualChooser::setSimpleConfiguration( void )
{
    clear();
    addAttribute( RGBA );
    addAttribute( DoubleBuffer );
    addAttribute( DepthSize, 16 );
}

void VisualChooser::clear() 
{
    _visual_attributes.clear();
    _vinfo = 0L;

    // Always use UseGL
    addAttribute( UseGL );
}


#ifdef _X11_IMPLEMENTATION
#include <GL/glx.h>


int VisualChooser::adaptAttribute( VisualAttribute &attr, VisualInfo *vinfo )
{
    switch( attr.attribute() )
    {
        case UseGL:         return GLX_USE_GL;
        case BufferSize:     return GLX_BUFFER_SIZE;
        case Level:          return GLX_LEVEL;
        case RGBA:           return GLX_RGBA;
        case DoubleBuffer:   return GLX_DOUBLEBUFFER;
        case Stereo:         return GLX_STEREO;
        case AuxBuffers:     return GLX_AUX_BUFFERS;
        case RedSize:        return GLX_RED_SIZE;
        case GreenSize:      return GLX_GREEN_SIZE;
        case BlueSize:       return GLX_BLUE_SIZE;
        case AlphaSize:      return GLX_ALPHA_SIZE;
        case DepthSize:      return GLX_DEPTH_SIZE;
        case StencilSize:    return GLX_STENCIL_SIZE;
        case AccumRedSize:   return GLX_ACCUM_RED_SIZE;
        case AccumGreenSize: return GLX_ACCUM_GREEN_SIZE;
        case AccumBlueSize:  return GLX_ACCUM_BLUE_SIZE;
        case AccumAlphaSize: return GLX_ACCUM_ALPHA_SIZE;
        default: return (int)attr.attribute();
    }
    return 0;
}



VisualInfo *VisualChooser::choose( Display *dpy, int screen, bool strict_adherence)
{
    // Visual info was previously forced.
    if( _vinfo != NULL )
        return _vinfo;

    if( _visual_id != 0 )
    {

        VisualInfo temp;
    long mask = VisualIDMask;
    int n;
    temp.visualid = _visual_id;
        _vinfo = XGetVisualInfo( dpy, mask, &temp, &n );

    if( _vinfo != NULL || strict_adherence )
        return _vinfo;

    }

    if( _visual_attributes.size() == 0 )
        setSimpleConfiguration();

    vector<VisualAttribute>::iterator p;
    vector<int>va;

    for( p = _visual_attributes.begin(); p != _visual_attributes.end(); p++ )
    {
    va.push_back( adaptAttribute((*p)));
     if( (*p).hasParameter() )
        va.push_back( (*p).parameter() );
    }
    va.push_back(0);

    if( strict_adherence )
    {
    _vinfo = glXChooseVisual( dpy, screen, &(va.front()) );
    }
    else
    {
        p = _visual_attributes.end() - 1;
    
        while( _vinfo == NULL && va.size() > 0)
        {
            _vinfo = glXChooseVisual( dpy, screen, &(va.front()) );
        if( _vinfo == NULL && va.size() > 0 )
        {
            
                // should we report an message that we're relaxing constraints here?
                // std::cout << "Popping attributes"<<std::endl;

            va.pop_back(); // Pop the NULL terminator
            if( (*p).hasParameter() && va.size() >= 2 )
            {
            va.pop_back();
            va.pop_back();
            }
            else
            va.pop_back();
            va.push_back(0); // Push a new NULL terminator
        if( p == _visual_attributes.begin() )
            break;
            p --;
        }
        }
    }

    _visual_id = _vinfo->visualid;

    return  _vinfo;
}
#endif

#ifdef _WIN32_IMPLEMENTATION
int VisualChooser::adaptAttribute( VisualAttribute &attr, VisualInfo *vinfo )
{
    switch (attr.attribute()) 
    {
        case UseGL:
              vinfo->dwFlags |= PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW;
              break;

        case BufferSize:
              vinfo->cColorBits = attr.parameter();
              break;
        
        case Level:
          /* the bReserved flag of the pfd contains the
         overlay/underlay info. */
              vinfo->bReserved = attr.parameter();
              break;
        case RGBA:
              vinfo->iPixelType = PFD_TYPE_RGBA;
              break;

        case DoubleBuffer:
              vinfo->dwFlags |= PFD_DOUBLEBUFFER  | PFD_SWAP_EXCHANGE;
              break;

        case Stereo:
              //stereo = 1;
              vinfo->dwFlags |= PFD_STEREO;
              break;

        case AuxBuffers:
              vinfo->cAuxBuffers = attr.parameter();
              break;

        case RedSize:
              vinfo->cRedBits = attr.parameter();
              break;

        case GreenSize:
              vinfo->cGreenBits = attr.parameter();
              break;

        case BlueSize:
              vinfo->cBlueBits = attr.parameter();
              break;

        case AlphaSize:
              vinfo->cAlphaBits = attr.parameter();
              break;

        case DepthSize:
              vinfo->cDepthBits = attr.parameter();
              break;

        case StencilSize:
              vinfo->cStencilBits = attr.parameter();
              break;

        case AccumRedSize:
        case AccumGreenSize:
        case AccumBlueSize:
        case AccumAlphaSize:
      /* I believe that WGL only used the cAccumRedBits,
         cAccumBlueBits, cAccumGreenBits, and cAccumAlphaBits fields
         when returning info about the accumulation buffer precision.
         Only cAccumBits is used for requesting an accumulation
         buffer. */
        vinfo->cAccumBits = 1;
            break;
    }

    return 0;
}


VisualInfo *VisualChooser::choose( Display *dpy, int screen, bool strict_adherence)
{
    screen = screen; 
    //static PIXELFORMATDESCRIPTOR pfd;
    if( _vinfo == NULL )
    _vinfo = new VisualInfo;

    /*
    if( _visual_id != 0 )
    return _visual_id;
    */

    if( _visual_attributes.size() == 0 )
        setSimpleConfiguration();

    
    memset(_vinfo, 0, sizeof(PIXELFORMATDESCRIPTOR));
    _vinfo->nSize = (sizeof(PIXELFORMATDESCRIPTOR));
    _vinfo->nVersion = 1;
    vector<VisualAttribute>::iterator p;
    for( p = _visual_attributes.begin(); p != _visual_attributes.end(); p++)
    adaptAttribute( (*p), _vinfo );

/*
    _vinfo->dwFlags = PFD_DRAW_TO_WINDOW | 
              PFD_SUPPORT_OPENGL | 
              PFD_DOUBLEBUFFER | PFD_SWAP_EXCHANGE;
    _vinfo->iPixelType = PFD_TYPE_RGBA;
    _vinfo->cColorBits = 24;
    _vinfo->cDepthBits = 16;
    _vinfo->cAccumBits = 0;
    _vinfo->cStencilBits = 0;
*/

    _visual_id = ChoosePixelFormat(*dpy, _vinfo );
    //if( strict_adherence )
    //;


    return  _vinfo;
}

#endif

unsigned int VisualChooser::getVisualID()  const
{
    return _visual_id;
}



